@@ -2,7 +2,7 @@
 #
 # This file is part of Nominatim. (https://nominatim.org)
 #
-# Copyright (C) 2024 by the Nominatim developer community.
+# Copyright (C) 2025 by the Nominatim developer community.
 # For a full list of authors see the git log.
 """
 Implementation of classes for API access via libraries.
@@ -14,6 +14,11 @@ import sys
 import contextlib
 from pathlib import Path
 
+if sys.version_info >= (3, 11):
+    from asyncio import timeout_at
+else:
+    from async_timeout import timeout_at
+
 import sqlalchemy as sa
 import sqlalchemy.ext.asyncio as sa_asyncio
 
@@ -26,6 +31,7 @@ from .connection import SearchConnection
 from .status import get_status, StatusResult
 from .lookup import get_places, get_detailed_place
 from .reverse import ReverseGeocoder
+from .timeout import Timeout
 from . import search as nsearch
 from . import types as ntyp
 from .results import DetailedResult, ReverseResult, SearchResults
@@ -62,6 +68,8 @@ class NominatimAPIAsync:
         self.config = Configuration(project_dir, environ)
         self.query_timeout = self.config.get_int('QUERY_TIMEOUT') \
             if self.config.QUERY_TIMEOUT else None
+        self.request_timeout = self.config.get_int('REQUEST_TIMEOUT') \
+            if self.config.REQUEST_TIMEOUT else None
         self.reverse_restrict_to_country_area = self.config.get_bool('SEARCH_WITHIN_COUNTRIES')
         self.server_version = 0
 
@@ -170,12 +178,15 @@ class NominatimAPIAsync:
         await self.close()
 
     @contextlib.asynccontextmanager
-    async def begin(self) -> AsyncIterator[SearchConnection]:
+    async def begin(self, abs_timeout: Optional[float] = None) -> AsyncIterator[SearchConnection]:
         """ Create a new connection with automatic transaction handling.
 
             This function may be used to get low-level access to the database.
             Refer to the documentation of SQLAlchemy for details how to use
             the connection object.
+
+            You may optionally give an absolute timeout until when to wait
+            for a connection to become available.
         """
         if self._engine is None:
             await self.setup_database()
@@ -183,14 +194,15 @@ class NominatimAPIAsync:
         assert self._engine is not None
         assert self._tables is not None
 
-        async with self._engine.begin() as conn:
+        async with timeout_at(abs_timeout), self._engine.begin() as conn:
             yield SearchConnection(conn, self._tables, self._property_cache, self.config)
 
     async def status(self) -> StatusResult:
         """ Return the status of the database.
         """
+        timeout = Timeout(self.request_timeout)
         try:
-            async with self.begin() as conn:
+            async with self.begin(abs_timeout=timeout.abs) as conn:
                 conn.set_query_timeout(self.query_timeout)
                 status = await get_status(conn)
         except (PGCORE_ERROR, sa.exc.OperationalError):
@@ -203,24 +215,30 @@ class NominatimAPIAsync:
 
             Returns None if there is no entry under the given ID.
         """
+        timeout = Timeout(self.request_timeout)
         details = ntyp.LookupDetails.from_kwargs(params)
-        async with self.begin() as conn:
-            conn.set_query_timeout(self.query_timeout)
-            if details.keywords:
-                await nsearch.make_query_analyzer(conn)
-            return await get_detailed_place(conn, place, details)
+        with details.query_stats as qs:
+            async with self.begin(abs_timeout=timeout.abs) as conn:
+                qs.log_time('start_query')
+                conn.set_query_timeout(self.query_timeout)
+                if details.keywords:
+                    await nsearch.make_query_analyzer(conn)
+                return await get_detailed_place(conn, place, details)
 
     async def lookup(self, places: Sequence[ntyp.PlaceRef], **params: Any) -> SearchResults:
         """ Get simple information about a list of places.
 
             Returns a list of place information for all IDs that were found.
         """
+        timeout = Timeout(self.request_timeout)
         details = ntyp.LookupDetails.from_kwargs(params)
-        async with self.begin() as conn:
-            conn.set_query_timeout(self.query_timeout)
-            if details.keywords:
-                await nsearch.make_query_analyzer(conn)
-            return await get_places(conn, places, details)
+        with details.query_stats as qs:
+            async with self.begin(abs_timeout=timeout.abs) as conn:
+                qs.log_time('start_query')
+                conn.set_query_timeout(self.query_timeout)
+                if details.keywords:
+                    await nsearch.make_query_analyzer(conn)
+                return await get_places(conn, places, details)
 
     async def reverse(self, coord: ntyp.AnyPoint, **params: Any) -> Optional[ReverseResult]:
         """ Find a place by its coordinates. Also known as reverse geocoding.
@@ -233,29 +251,34 @@ class NominatimAPIAsync:
             # There are no results to be expected outside valid coordinates.
             return None
 
+        timeout = Timeout(self.request_timeout)
         details = ntyp.ReverseDetails.from_kwargs(params)
-        async with self.begin() as conn:
-            conn.set_query_timeout(self.query_timeout)
-            if details.keywords:
-                await nsearch.make_query_analyzer(conn)
-            geocoder = ReverseGeocoder(conn, details,
-                                       self.reverse_restrict_to_country_area)
-            return await geocoder.lookup(coord)
+        with details.query_stats as qs:
+            async with self.begin(abs_timeout=timeout.abs) as conn:
+                qs.log_time('start_query')
+                conn.set_query_timeout(self.query_timeout)
+                if details.keywords:
+                    await nsearch.make_query_analyzer(conn)
+                geocoder = ReverseGeocoder(conn, details,
+                                           self.reverse_restrict_to_country_area)
+                return await geocoder.lookup(coord)
 
     async def search(self, query: str, **params: Any) -> SearchResults:
         """ Find a place by free-text search. Also known as forward geocoding.
         """
-        query = query.strip()
-        if not query:
-            raise UsageError('Nothing to search for.')
+        timeout = Timeout(self.request_timeout)
+        details = ntyp.SearchDetails.from_kwargs(params)
+        with details.query_stats as qs:
+            query = query.strip()
+            if not query:
+                raise UsageError('Nothing to search for.')
 
-        async with self.begin() as conn:
-            conn.set_query_timeout(self.query_timeout)
-            geocoder = nsearch.ForwardGeocoder(conn, ntyp.SearchDetails.from_kwargs(params),
-                                               self.config.get_int('REQUEST_TIMEOUT')
-                                               if self.config.REQUEST_TIMEOUT else None)
-            phrases = [nsearch.Phrase(nsearch.PHRASE_ANY, p.strip()) for p in query.split(',')]
-            return await geocoder.lookup(phrases)
+            async with self.begin(abs_timeout=timeout.abs) as conn:
+                qs.log_time('start_query')
+                conn.set_query_timeout(self.query_timeout)
+                geocoder = nsearch.ForwardGeocoder(conn, details, timeout)
+                phrases = [nsearch.Phrase(nsearch.PHRASE_ANY, p.strip()) for p in query.split(',')]
+                return await geocoder.lookup(phrases)
 
     async def search_address(self, amenity: Optional[str] = None,
                              street: Optional[str] = None,
@@ -267,10 +290,9 @@ class NominatimAPIAsync:
                              **params: Any) -> SearchResults:
         """ Find an address using structured search.
         """
-        async with self.begin() as conn:
-            conn.set_query_timeout(self.query_timeout)
-            details = ntyp.SearchDetails.from_kwargs(params)
-
+        timeout = Timeout(self.request_timeout)
+        details = ntyp.SearchDetails.from_kwargs(params)
+        with details.query_stats as qs:
             phrases: List[nsearch.Phrase] = []
 
             if amenity:
@@ -309,10 +331,11 @@ class NominatimAPIAsync:
                 if amenity:
                     details.layers |= ntyp.DataLayer.POI
 
-            geocoder = nsearch.ForwardGeocoder(conn, details,
-                                               self.config.get_int('REQUEST_TIMEOUT')
-                                               if self.config.REQUEST_TIMEOUT else None)
-            return await geocoder.lookup(phrases)
+            async with self.begin(abs_timeout=timeout.abs) as conn:
+                qs.log_time('start_query')
+                conn.set_query_timeout(self.query_timeout)
+                geocoder = nsearch.ForwardGeocoder(conn, details, timeout)
+                return await geocoder.lookup(phrases)
 
     async def search_category(self, categories: List[Tuple[str, str]],
                               near_query: Optional[str] = None,
@@ -321,23 +344,24 @@ class NominatimAPIAsync:
             The near place may either be given as an unstructured search
             query in itself or as coordinates.
         """
-        if not categories:
-            return SearchResults()
-
+        timeout = Timeout(self.request_timeout)
         details = ntyp.SearchDetails.from_kwargs(params)
-        async with self.begin() as conn:
-            conn.set_query_timeout(self.query_timeout)
-            if near_query:
-                phrases = [nsearch.Phrase(nsearch.PHRASE_ANY, p) for p in near_query.split(',')]
-            else:
-                phrases = []
-                if details.keywords:
-                    await nsearch.make_query_analyzer(conn)
+        with details.query_stats as qs:
+            if not categories:
+                return SearchResults()
 
-            geocoder = nsearch.ForwardGeocoder(conn, details,
-                                               self.config.get_int('REQUEST_TIMEOUT')
-                                               if self.config.REQUEST_TIMEOUT else None)
-            return await geocoder.lookup_pois(categories, phrases)
+            async with self.begin(abs_timeout=timeout.abs) as conn:
+                qs.log_time('start_query')
+                conn.set_query_timeout(self.query_timeout)
+                if near_query:
+                    phrases = [nsearch.Phrase(nsearch.PHRASE_ANY, p) for p in near_query.split(',')]
+                else:
+                    phrases = []
+                    if details.keywords:
+                        await nsearch.make_query_analyzer(conn)
+
+                geocoder = nsearch.ForwardGeocoder(conn, details, timeout)
+                return await geocoder.lookup_pois(categories, phrases)
 
 
 class NominatimAPI:
@@ -434,6 +458,8 @@ class NominatimAPI:
                 Only POI places can have parents. (Default: False)
               keywords (bool): Add detailed information about the search terms
                 used for this place.
+              query_stats (QueryStatistics): When given collects statistics
+                about the query execution.
 
             Returns:
               source_table (enum): Data source of the place. See below for possible values.
@@ -516,6 +542,8 @@ class NominatimAPI:
                 Only POI places can have parents. (Default: False)
               keywords (bool): Add detailed information about the search terms
                 used for this place.
+              query_stats (QueryStatistics): When given collects statistics
+                about the query execution.
 
             Returns:
               source_table (enum): Data source of the place. See below for possible values.
@@ -596,6 +624,8 @@ class NominatimAPI:
                 Only POI places can have parents. (Default: False)
               keywords (bool): Add detailed information about the search terms
                 used for this place.
+              query_stats (QueryStatistics): When given collects statistics
+                about the query execution.
 
             Returns:
               source_table (enum): Data source of the place. See below for possible values.
@@ -695,6 +725,8 @@ class NominatimAPI:
                 Only POI places can have parents. (Default: False)
               keywords (bool): Add detailed information about the search terms
                 used for this place.
+              query_stats (QueryStatistics): When given collects statistics
+                about the query execution.
 
             Returns:
               source_table (enum): Data source of the place. See below for possible values.
@@ -811,6 +843,8 @@ class NominatimAPI:
                 Only POI places can have parents. (Default: False)
               keywords (bool): Add detailed information about the search terms
                 used for this place.
+              query_stats (QueryStatistics): When given collects statistics
+                about the query execution.
 
             Returns:
               source_table (enum): Data source of the place. See below for possible values.
@@ -918,6 +952,8 @@ class NominatimAPI:
                 Only POI places can have parents. (Default: False)
               keywords (bool): Add detailed information about the search terms
                 used for this place.
+              query_stats (QueryStatistics): When given collects statistics
+                about the query execution.
 
             Returns:
               source_table (enum): Data source of the place. See below for possible values.
