@@ -2,7 +2,7 @@
 #
 # This file is part of Nominatim. (https://nominatim.org)
 #
-# Copyright (C) 2024 by the Nominatim developer community.
+# Copyright (C) 2025 by the Nominatim developer community.
 # For a full list of authors see the git log.
 """
 Conversion from token assignment to an abstract DB search.
@@ -146,7 +146,7 @@ class SearchBuilder:
             if address:
                 sdata.lookups = [dbf.FieldLookup('nameaddress_vector',
                                                  [t.token for r in address
-                                                  for t in self.query.get_partials_list(r)],
+                                                  for t in self.query.iter_partials(r)],
                                                  lookups.Restrict)]
             yield dbs.PostcodeSearch(penalty, sdata)
 
@@ -155,32 +155,39 @@ class SearchBuilder:
         """ Build a simple address search for special entries where the
             housenumber is the main name token.
         """
-        sdata.lookups = [dbf.FieldLookup('name_vector', [t.token for t in hnrs], lookups.LookupAny)]
-        expected_count = sum(t.count for t in hnrs)
-
-        partials = {t.token: t.addr_count for trange in address
-                    for t in self.query.get_partials_list(trange)}
+        partials = dbf.CountedTokenIDs((t for trange in address
+                                        for t in self.query.iter_partials(trange)),
+                                       'addr_count')
 
         if not partials:
             # can happen when none of the partials is indexed
             return
 
-        if expected_count < 8000:
-            sdata.lookups.append(dbf.FieldLookup('nameaddress_vector',
-                                                 list(partials), lookups.Restrict))
-        elif len(partials) != 1 or list(partials.values())[0] < 10000:
-            sdata.lookups.append(dbf.FieldLookup('nameaddress_vector',
-                                                 list(partials), lookups.LookupAll))
+        expected_count = sum(t.count for t in hnrs)
+        hnr_tokens = [t.token for t in hnrs]
+
+        if expected_count < 10000:
+            sdata.lookups = [dbf.FieldLookup('name_vector', hnr_tokens, lookups.LookupAny),
+                             dbf.FieldLookup('nameaddress_vector',
+                                             partials.get_tokens(),
+                                             lookups.Restrict)]
         else:
-            addr_fulls = [t.token for t
-                          in self.query.get_tokens(address[0], qmod.TOKEN_WORD)]
-            if len(addr_fulls) > 5:
-                return
-            sdata.lookups.append(
-                dbf.FieldLookup('nameaddress_vector', addr_fulls, lookups.LookupAny))
+            split = partials.get_num_lookup_tokens(20000, 5)
+            if split > 0:
+                sdata.lookups = partials.split_lookup(split, 'nameaddress_vector')
+                sdata.lookups.append(
+                    dbf.FieldLookup('name_vector', hnr_tokens, lookups.Restrict))
+            else:
+                addr_fulls = [t.token for t in
+                              self.query.get_tokens(address[0], qmod.TOKEN_WORD)]
+                if len(addr_fulls) > 5:
+                    return
+                sdata.lookups = [
+                    dbf.FieldLookup('name_vector', hnr_tokens, lookups.LookupAny),
+                    dbf.FieldLookup('nameaddress_vector', addr_fulls, lookups.LookupAny)]
 
         sdata.housenumbers = dbf.WeightedStrings([], [])
-        yield dbs.PlaceSearch(0.05, sdata, expected_count)
+        yield dbs.PlaceSearch(0.0, sdata, expected_count, True)
 
     def build_name_search(self, sdata: dbf.SearchData,
                           name: qmod.TokenRange, address: List[qmod.TokenRange],
@@ -194,7 +201,10 @@ class SearchBuilder:
                 sdata.rankings.append(ranking)
             for penalty, count, lookup in self.yield_lookups(name, address):
                 sdata.lookups = lookup
-                yield dbs.PlaceSearch(penalty + name_penalty, sdata, count)
+                if sdata.housenumbers:
+                    yield dbs.AddressSearch(penalty + name_penalty, sdata, count, bool(address))
+                else:
+                    yield dbs.PlaceSearch(penalty + name_penalty, sdata, count, bool(address))
 
     def yield_lookups(self, name: qmod.TokenRange, address: List[qmod.TokenRange]
                       ) -> Iterator[Tuple[float, int, List[dbf.FieldLookup]]]:
@@ -202,88 +212,102 @@ class SearchBuilder:
             be searched for. This takes into account how frequent the terms
             are and tries to find a lookup that optimizes index use.
         """
-        penalty = 0.0  # extra penalty
-        name_partials = {t.token: t for t in self.query.get_partials_list(name)}
+        name_partials = dbf.CountedTokenIDs(self.query.iter_partials(name))
+        addr_partials = dbf.CountedTokenIDs((t for r in address
+                                             for t in self.query.iter_partials(r)),
+                                            'addr_count')
 
-        addr_partials = [t for r in address for t in self.query.get_partials_list(r)]
-        addr_tokens = list({t.token for t in addr_partials})
-
-        exp_count = min(t.count for t in name_partials.values()) / (3**(len(name_partials) - 1))
-
-        if (len(name_partials) > 3 or exp_count < 8000):
-            yield penalty, exp_count, dbf.lookup_by_names(list(name_partials.keys()), addr_tokens)
-            return
-
-        addr_count = min(t.addr_count for t in addr_partials) if addr_partials else 50000
-        # Partial term to frequent. Try looking up by rare full names first.
-        name_fulls = self.query.get_tokens(name, qmod.TOKEN_WORD)
-        if name_fulls:
-            fulls_count = sum(t.count for t in name_fulls)
-
-            if fulls_count < 50000 or addr_count < 50000:
-                yield penalty, fulls_count / (2**len(addr_tokens)), \
-                    self.get_full_name_ranking(name_fulls, addr_partials,
-                                               fulls_count > 30000 / max(1, len(addr_tokens)))
-
-        # To catch remaining results, lookup by name and address
-        # We only do this if there is a reasonable number of results expected.
-        exp_count /= 2**len(addr_tokens)
-        if exp_count < 10000 and addr_count < 20000:
-            penalty += 0.35 * max(1 if name_fulls else 0.1,
-                                  5 - len(name_partials) - len(addr_tokens))
-            yield penalty, exp_count, \
-                self.get_name_address_ranking(list(name_partials.keys()), addr_partials)
-
-    def get_name_address_ranking(self, name_tokens: List[int],
-                                 addr_partials: List[qmod.Token]) -> List[dbf.FieldLookup]:
-        """ Create a ranking expression looking up by name and address.
-        """
-        lookup = [dbf.FieldLookup('name_vector', name_tokens, lookups.LookupAll)]
-
-        addr_restrict_tokens = []
-        addr_lookup_tokens = []
-        for t in addr_partials:
-            if t.addr_count > 20000:
-                addr_restrict_tokens.append(t.token)
-            else:
-                addr_lookup_tokens.append(t.token)
-
-        if addr_restrict_tokens:
-            lookup.append(dbf.FieldLookup('nameaddress_vector',
-                                          addr_restrict_tokens, lookups.Restrict))
-        if addr_lookup_tokens:
-            lookup.append(dbf.FieldLookup('nameaddress_vector',
-                                          addr_lookup_tokens, lookups.LookupAll))
-
-        return lookup
-
-    def get_full_name_ranking(self, name_fulls: List[qmod.Token], addr_partials: List[qmod.Token],
-                              use_lookup: bool) -> List[dbf.FieldLookup]:
-        """ Create a ranking expression with full name terms and
-            additional address lookup. When 'use_lookup' is true, then
-            address lookups will use the index, when the occurrences are not
-            too many.
-        """
-        if use_lookup:
-            addr_restrict_tokens = []
-            addr_lookup_tokens = [t.token for t in addr_partials]
+        if not addr_partials:
+            yield from self.yield_name_only_lookups(name_partials, name)
         else:
-            addr_restrict_tokens = [t.token for t in addr_partials]
-            addr_lookup_tokens = []
+            yield from self.yield_address_lookups(name_partials, addr_partials, name)
 
-        return dbf.lookup_by_any_name([t.token for t in name_fulls],
-                                      addr_restrict_tokens, addr_lookup_tokens)
+    def yield_name_only_lookups(self, partials: dbf.CountedTokenIDs, name: qmod.TokenRange
+                                ) -> Iterator[Tuple[float, int, List[dbf.FieldLookup]]]:
+        """ Yield the best lookup for a name-only search.
+        """
+        split = partials.get_num_lookup_tokens(30000, 6)
+
+        if split > 0:
+            yield 0.0, partials.expected_for_all_search(5), \
+                  partials.split_lookup(split, 'name_vector')
+        else:
+            # lots of results expected: try lookup by full names first
+            name_fulls = list(filter(lambda t: t.count < 50000,
+                                     self.query.get_tokens(name, qmod.TOKEN_WORD)))
+            if name_fulls:
+                yield 0.0, sum(t.count for t in name_fulls), \
+                      dbf.lookup_by_any_name([t.token for t in name_fulls], [], [])
+
+            # look the name up by its partials
+            exp_count = partials.expected_for_all_search(5)
+            if exp_count < 50000:
+                yield 1.0, exp_count, \
+                      [dbf.FieldLookup('name_vector', partials.get_tokens(), lookups.LookupAll)]
+
+    def yield_address_lookups(self, name_partials: dbf.CountedTokenIDs,
+                              addr_partials: dbf.CountedTokenIDs, name: qmod.TokenRange,
+                              ) -> Iterator[Tuple[float, int, List[dbf.FieldLookup]]]:
+        penalty = 0.0  # extra penalty
+
+        name_split = name_partials.get_num_lookup_tokens(20000, 6)
+        addr_split = addr_partials.get_num_lookup_tokens(10000, 3)
+
+        if name_split < 0 and addr_split < 0:
+            # Partial term too frequent. Try looking up by rare full names first.
+            name_fulls = self.query.get_tokens(name, qmod.TOKEN_WORD)
+            if name_fulls:
+                fulls_count = sum(t.count for t in name_fulls)
+
+                if fulls_count < 80000:
+                    yield 0.0, fulls_count, \
+                          dbf.lookup_by_any_name([t.token for t in name_fulls],
+                                                 addr_partials.get_tokens(),
+                                                 [])
+                    penalty += 0.2
+            penalty += 0.4
+
+            name_split = name_partials.get_num_lookup_tokens(50000, 10)
+            addr_split = addr_partials.get_num_lookup_tokens(30000, 5)
+
+        if name_split > 0 \
+           and (addr_split < 0 or name_partials.min_count() <= addr_partials.min_count()):
+            # lookup by name
+            lookup = name_partials.split_lookup(name_split, 'name_vector')
+            lookup.append(dbf.FieldLookup('nameaddress_vector',
+                                          addr_partials.get_tokens(), lookups.Restrict))
+            yield penalty, name_partials.expected_for_all_search(5), lookup
+        elif addr_split > 0:
+            # lookup by address
+            lookup = addr_partials.split_lookup(addr_split, 'nameaddress_vector')
+            lookup.append(dbf.FieldLookup('name_vector',
+                                          name_partials.get_tokens(), lookups.Restrict))
+            yield penalty, addr_partials.expected_for_all_search(3), lookup
+        elif len(name_partials) > 1:
+            penalty += 0.5
+            # To catch remaining results, lookup by name and address
+            # We only do this if there is a reasonable number of results expected.
+            exp_count = min(name_partials.min_count(), addr_partials.min_count())
+            exp_count = int(exp_count / (min(3, len(name_partials)) + min(3, len(addr_partials))))
+            if exp_count < 50000:
+                lookup = name_partials.split_lookup(3, 'name_vector')
+                lookup.extend(addr_partials.split_lookup(3, 'nameaddress_vector'))
+
+                yield penalty, exp_count, lookup
 
     def get_name_ranking(self, trange: qmod.TokenRange,
                          db_field: str = 'name_vector') -> dbf.FieldRanking:
         """ Create a ranking expression for a name term in the given range.
         """
         name_fulls = self.query.get_tokens(trange, qmod.TOKEN_WORD)
-        ranks = [dbf.RankedTokens(t.penalty, [t.token]) for t in name_fulls]
+        full_word_penalty = self.query.get_in_word_penalty(trange)
+        ranks = [dbf.RankedTokens(t.penalty + full_word_penalty, [t.token])
+                 for t in name_fulls]
         ranks.sort(key=lambda r: r.penalty)
         # Fallback, sum of penalty for partials
-        name_partials = self.query.get_partials_list(trange)
-        default = sum(t.penalty for t in name_partials) + 0.2
+        default = sum(t.penalty for t in self.query.iter_partials(trange)) + 0.2
+        default += sum(n.word_break_penalty
+                       for n in self.query.nodes[trange.start + 1:trange.end])
         return dbf.FieldRanking(db_field, default, ranks)
 
     def get_addr_ranking(self, trange: qmod.TokenRange) -> dbf.FieldRanking:
@@ -295,36 +319,43 @@ class SearchBuilder:
         ranks: List[dbf.RankedTokens] = []
 
         while todo:
-            neglen, pos, rank = heapq.heappop(todo)
+            _, pos, rank = heapq.heappop(todo)
+            # partial node
+            partial = self.query.nodes[pos].partial
+            if partial is not None:
+                if pos + 1 < trange.end:
+                    penalty = rank.penalty + partial.penalty \
+                              + self.query.nodes[pos + 1].word_break_penalty
+                    heapq.heappush(todo, (-(pos + 1), pos + 1,
+                                   dbf.RankedTokens(penalty, rank.tokens)))
+                else:
+                    ranks.append(dbf.RankedTokens(rank.penalty + partial.penalty,
+                                                  rank.tokens))
+            # full words
             for tlist in self.query.nodes[pos].starting:
-                if tlist.ttype in (qmod.TOKEN_PARTIAL, qmod.TOKEN_WORD):
+                if tlist.ttype == qmod.TOKEN_WORD:
                     if tlist.end < trange.end:
-                        chgpenalty = PENALTY_WORDCHANGE[self.query.nodes[tlist.end].btype]
-                        if tlist.ttype == qmod.TOKEN_PARTIAL:
-                            penalty = rank.penalty + chgpenalty \
-                                      + max(t.penalty for t in tlist.tokens)
-                            heapq.heappush(todo, (neglen - 1, tlist.end,
-                                                  dbf.RankedTokens(penalty, rank.tokens)))
-                        else:
-                            for t in tlist.tokens:
-                                heapq.heappush(todo, (neglen - 1, tlist.end,
-                                                      rank.with_token(t, chgpenalty)))
+                        chgpenalty = self.query.nodes[tlist.end].word_break_penalty \
+                                     + self.query.get_in_word_penalty(
+                                            qmod.TokenRange(pos, tlist.end))
+                        for t in tlist.tokens:
+                            heapq.heappush(todo, (-tlist.end, tlist.end,
+                                                  rank.with_token(t, chgpenalty)))
                     elif tlist.end == trange.end:
-                        if tlist.ttype == qmod.TOKEN_PARTIAL:
-                            ranks.append(dbf.RankedTokens(rank.penalty
-                                                          + max(t.penalty for t in tlist.tokens),
-                                                          rank.tokens))
-                        else:
-                            ranks.extend(rank.with_token(t, 0.0) for t in tlist.tokens)
-                        if len(ranks) >= 10:
-                            # Too many variants, bail out and only add
-                            # Worst-case Fallback: sum of penalty of partials
-                            name_partials = self.query.get_partials_list(trange)
-                            default = sum(t.penalty for t in name_partials) + 0.2
-                            ranks.append(dbf.RankedTokens(rank.penalty + default, []))
-                            # Bail out of outer loop
-                            todo.clear()
-                            break
+                        chgpenalty = self.query.get_in_word_penalty(
+                                        qmod.TokenRange(pos, tlist.end))
+                        ranks.extend(rank.with_token(t, chgpenalty)
+                                     for t in tlist.tokens)
+
+            if len(ranks) >= 10:
+                # Too many variants, bail out and only add
+                # Worst-case Fallback: sum of penalty of partials
+                default = sum(t.penalty for t in self.query.iter_partials(trange)) + 0.2
+                default += sum(n.word_break_penalty
+                               for n in self.query.nodes[trange.start + 1:trange.end])
+                ranks.append(dbf.RankedTokens(rank.penalty + default, []))
+                # Bail out of outer loop
+                break
 
         ranks.sort(key=lambda r: len(r.tokens))
         default = ranks[0].penalty + 0.3
@@ -344,6 +375,7 @@ class SearchBuilder:
             if not tokens:
                 return None
             sdata.set_strings('countries', tokens)
+            sdata.penalty += self.query.get_in_word_penalty(assignment.country)
         elif self.details.countries:
             sdata.countries = dbf.WeightedStrings(self.details.countries,
                                                   [0.0] * len(self.details.countries))
@@ -351,29 +383,24 @@ class SearchBuilder:
             sdata.set_strings('housenumbers',
                               self.query.get_tokens(assignment.housenumber,
                                                     qmod.TOKEN_HOUSENUMBER))
+            sdata.penalty += self.query.get_in_word_penalty(assignment.housenumber)
         if assignment.postcode:
             sdata.set_strings('postcodes',
                               self.query.get_tokens(assignment.postcode,
                                                     qmod.TOKEN_POSTCODE))
+            sdata.penalty += self.query.get_in_word_penalty(assignment.postcode)
         if assignment.qualifier:
             tokens = self.get_qualifier_tokens(assignment.qualifier)
             if not tokens:
                 return None
             sdata.set_qualifiers(tokens)
+            sdata.penalty += self.query.get_in_word_penalty(assignment.qualifier)
         elif self.details.categories:
             sdata.qualifiers = dbf.WeightedCategories(self.details.categories,
                                                       [0.0] * len(self.details.categories))
 
         if assignment.address:
-            if not assignment.name and assignment.housenumber:
-                # housenumber search: the first item needs to be handled like
-                # a name in ranking or penalties are not comparable with
-                # normal searches.
-                sdata.set_ranking([self.get_name_ranking(assignment.address[0],
-                                                         db_field='nameaddress_vector')]
-                                  + [self.get_addr_ranking(r) for r in assignment.address[1:]])
-            else:
-                sdata.set_ranking([self.get_addr_ranking(r) for r in assignment.address])
+            sdata.set_ranking([self.get_addr_ranking(r) for r in assignment.address])
         else:
             sdata.rankings = []
 
@@ -419,14 +446,3 @@ class SearchBuilder:
             return dbf.WeightedCategories(list(tokens.keys()), list(tokens.values()))
 
         return None
-
-
-PENALTY_WORDCHANGE = {
-    qmod.BREAK_START: 0.0,
-    qmod.BREAK_END: 0.0,
-    qmod.BREAK_PHRASE: 0.0,
-    qmod.BREAK_SOFT_PHRASE: 0.0,
-    qmod.BREAK_WORD: 0.1,
-    qmod.BREAK_PART: 0.2,
-    qmod.BREAK_TOKEN: 0.4
-}
